/******************************************************************************
* To the extent possible under law, Sebastian Roith
* <https://www.roith-unterrichtet.de> has waived all copyright and related or
* neighboring rights to this script. This work is published from Germany.
******************************************************************************/

AudioFeedbackTask.caching.push('bottom.svg');
AudioFeedbackTask.caching.push('delete.svg');
AudioFeedbackTask.caching.push('down.svg');
AudioFeedbackTask.caching.push('Go-previous.svg');
AudioFeedbackTask.caching.push('not-ok.svg');
AudioFeedbackTask.caching.push('ok.svg');
AudioFeedbackTask.caching.push('top.svg');
AudioFeedbackTask.caching.push('up.svg');
AudioFeedbackTask.caching.push('unknown.svg');

/**
* Klasse, die den Spielablauf eines Spiels, bei denen Textbausteine in die richtige Reihenfolge zu bringen sind, verwaltet, die Kommunikation mit der Lernplattform über SCORM-API übernimmt und die Anzeige der nötigen Elemente sicherstellt
*/
function Reihenfolgenspiel(
    nodeParent,     // DOM-Knoten, als dessen einzige Kinder das Spiel angezeigt werden soll
    objData         // Spieldaten
) {
    Reihenfolgenspiel.parent.call(this);
    this.data = objData;
    this.msgPlace = new ScoTools.MessagePlace();
    this.msgPlaceGeneral = new ScoTools.MessagePlace(true);
    this.msgPlaceGeneral.insertThisBefore(nodeParent);
    this.nodeParent = nodeParent;

    this.connect();
    this.establish();
    AudioFeedbackTask.cache();
    this.ready = true;
}
Reihenfolgenspiel.prototype = Object.create(AudioFeedbackTask.prototype);
Reihenfolgenspiel.prototype.constructor = Reihenfolgenspiel;
Reihenfolgenspiel.parent = AudioFeedbackTask;

// Ist das Spiel gestartet?
Reihenfolgenspiel.prototype.ready = false;

/**
* Als letztes Kind von this.nodeParent einen oder mehrere Absätze hinzufügen (sofern ein Text definiert ist)
* @return HTMLParagraphElement      Erster eingefügter Paragraph
*/
Reihenfolgenspiel.prototype._addParagraphs = function(
    key         // Schlüssel in this.data, unter dem evtl. die Absätze zeilenvorschubsepariert gespeichert sind
) {
    var self = this;
    var r = null;
    if ((typeof this.data[key] == 'string') || (this.data[key] instanceof String))
        this.data[key].split("\n").map(function(s) {
            var p = ScoTools.createHtml('p', self.nodeParent, null, s);
            if (!r)
                r = p;
        });
    return r;
}

/**
* Als letztes Kind von this.nodeParent eine Überschrift hinzufügen (sofern this.data[key] ein String ist)
* @return undefined
*/
Reihenfolgenspiel.prototype._addHeader = function(
    key,        // Schlüssel in this.data, unter dem der Überschriftentext gespeichert ist
    level       // Optional: Überschriftenordnung (Standardwert: 2)
) {
    if ((typeof this.data[key] == 'string') || (this.data[key] instanceof String))
        ScoTools.createHtml('h' + (level ? Math.floor(Math.min(6, Math.max(1, level))) : 2), this.nodeParent, null, this.data[key]);
}

// Schaltfläche zum Überprüfen oder zurücksetzen
Reihenfolgenspiel.prototype._btnSubmit = null;

// Führt das Drücken der Schaltfläche zum Neustart (true) oder zum Überprüfen bzw. Weiterbearbeiten (false)
Object.defineProperty(Reihenfolgenspiel.prototype, 'needRestart', {
    enumerable: true,
    set(v) {
        if (!this._btnSubmit)
            return;
        if (!v && this.needRestart) {
            var key = this.getDone() ? 'redoing' : 'task';
            if (key && this.data[key])
                this.showMsg(this.data[key], ScoTools.MSGTYPE_INFO);
        }
        if (v && this.data.labelRestart)
            this._setSubmitLabel(this.data.labelRestart);
        else if (this.data.labelCheck)
            this._setSubmitLabel(this.data.labelCheck);
        else
            this._setSubmitLabel('Überprüfen');
        this.listAnswere.update();
        this.listAvailable.update();
        this.write('lessonLocation', v ? 1 : 0);
        if (this.ready) {
            try {
                this._btnSubmit.animate([ { "transform": "scale(1)" }, { "transform": "scale(0)" }, { "transform": "scale(1)" } ], { "duration": 400 });
            } catch {}
        }
    },
    get() {
        return (!this._btnSubmit || (this.data.labelRestart && (this._btnSubmit.value == this.data.labelRestart)));
    }
});

/**
* Schaltflächentext auf label ändern und Wechsel ggf. animieren
* @return undefined
*/
Reihenfolgenspiel.prototype._setSubmitLabel = function(
    label       // Neue Beschriftung
) {
    this._btnSubmit.value = label;
    if (this.ready) {
        try {
            this._btnSubmit.animate([ { "transform": "scale(1)" }, { "transform": "scale(0)" }, { "transform": "scale(1)" } ], { "duration": 400 });
        } catch(ex) {}
    }
}

/**
* Darauf reagieren, wenn auf den Überprüfen-/Neustarten-Button gedrückt wird
* @return undefined
*/
Reihenfolgenspiel.prototype._procSubmit = function() {
    if (this._needClear) {
        this.procChange(null);
    } else if (this.needRestart) {
        this.needRestart = false;
        this.listAnswere.clear();
        this.listAvailable.clear();
        this.listAvailable.addArray(this._getIndizesAvailable(true));
    } else
        this.isCorrect();
}

// Liste mit ausgewählter Erklärung
Reihenfolgenspiel.prototype.listAnswere = null;

// Liste mit (noch) verfügbaren Textbausteinen
Reihenfolgenspiel.prototype.listAvailable = null;

/**
* Ermittelt die gespeicherte Liste an Textbausteinen
* @return Array     Indizes der gespeicherten Textbausteine
*/
Reihenfolgenspiel.prototype._getSaved = function() {
    if (this.isConnected()) {
        var data = this.read('suspendData');
        var maxIndex = this.data.parts.length - 1;
        if (data)
            return data.split("|").map(function(s){ return +s; }).filter(function(v) {
                return (Number.isFinite(v) && (Math.floor(Math.min(maxIndex, Math.max(0, v))) == v));
            });
    }
    return [];
}

/**
* Indizes der noch verfügbaren Textbausteine ermitteln
* @return Array     Indizes der noch verfügbaren Textbausteine
*/
Reihenfolgenspiel.prototype._getIndizesAvailable = function(
    boolUseAll      // Optional: Ignorieren, welche Textbausteine schon in Verwendung sind
) {
    var arr = [];
    var used = boolUseAll ? [] : this._getSaved();
    for (var n = 0; n < this.data.parts.length; ++n) {
        if (this.data.multiple || (used.indexOf(n) == -1))
            arr.push(n);
    }
    if (this.data.shuffle)
        arr = arr.map(function(value) {
            return { "value": value, "pos": Math.floor(Math.random() * 9007199254740991) };
        }).sort(function(a, b) {
            return a.pos - b.pos;
        }).map(function(o) {
            return o.value;
        });
    return arr;
}

/**
* Bisherigen Inhalt von this.nodeParent durch Spielfeld ersetzen
* @return undefined
*/
Reihenfolgenspiel.prototype.establish = function() {
    Reihenfolgenspiel.parent.prototype.establish.call(this);
    var mode = (this.isConnected()) ? this.lmslink.lessonMode : 'normal';
    if (this.nodeParent.firstChild)
        this.nodeParent.innerHTML = '';
    this._addHeader('labelInfo');
    this._addParagraphs('textInfo');
    this._addHeader('labelAnswere');
    var nodeBeforeMsg = this._addParagraphs('textAnswere');
    this.listAnswere = new Liste(this, true, this.getDone() ? 'emptyRedo' : 'emptySelection', this._getSaved());
    this.msgPlace.insertThisBefore(nodeBeforeMsg ? nodeBeforeMsg : this.listAnswere.node);
    if (mode != 'browse') {
        var form = ScoTools.createHtml('form', this.nodeParent, { "action": location.href, "method": "post" });
        var self = this;
        form.addEventListener('submit', function(ev) {
            ev.preventDefault();
            ev.stopPropagation();
            self._procSubmit();
        });
        this._btnSubmit = ScoTools.createHtml('input', form, { "type": "submit" });
    }
    this._addHeader('labelParts', 3);
    this._addParagraphs('textParts');
    this.listAvailable = new Liste(this, false, 'empty', this._getIndizesAvailable());
    var key = false;
    switch (mode) {
        case 'normal':
            key = this.getDone() ? 'redoing' : 'task';
            if (this.read('lessonLocation'))
                this.isCorrect();
            break;
        case 'browse':
            key = 'taskBrowse';
            break;
        case 'review':
            this.isCorrect();
            break;
    }
    if (key && this.data[key])
        this.showMsg(this.data[key], ScoTools.MSGTYPE_INFO);
    this.needRestart = (mode != 'normal');
}

/**
* Einer Liste mitteilen, ob sie editierbar ist
* @return String    "never" → Liste ist grundsätzlich nicht editierbar, "no" → Liste aktuell nicht editierbar, "yes" → Liste aktuell editierbar (ohne objListe: prinzipiell Editieren erlabut?)
*/
Reihenfolgenspiel.prototype.getListEditable = function(
    objListe    // Optional: Fragliche Liste-Instanz
) {
    if (!(objListe instanceof Liste) || (objListe == this.listAnswere)) {
        return this.needRestart ? "no" : "yes";
    } else
        return "never";
}

/**
* Darauf reagieren, dass der Listenpunkt objListenpunkt angeklickt wurde
* @return Boolean       Konnte keine Standardaktion durchgeführt werden?
*/
Reihenfolgenspiel.prototype.procClick = function(
    objListenpunkt      // Angeklickter Listenpunkt
) {
    if ((objListenpunkt.liste == this.listAvailable) && !this.needRestart) {
        var clone = this.listAnswere.add(objListenpunkt.index);
        clone.animate(objListenpunkt.getTop());
        if (!this.data.multiple)
            objListenpunkt.remove();
        return false;
    }
    return true;
}

/**
* Darauf reagieren, wenn eine Liste eine Veränderung meldet
* @return undefined
*/
Reihenfolgenspiel.prototype.procChange = function(
    objListe        // Liste, bei der sich was geändert hat
) {
    if ((objListe == this.listAnswere) && this.canSave()) {
        this.write('suspendData', this.listAnswere.getIndizes().join('|'));
        this.reportTime();
    }
    if (this._needClear) {
        this.msgPlace.clear();
        this.listAnswere.update();
        this._setSubmitLabel(this.data.labelCheck ? this.data.labelCheck : 'Überprüfen');
        this._needClear = false;
    }
}

/**
* Den Listenpunkt objListenpunkt aus seiner Liste entfernen
* @return undefined
*/
Reihenfolgenspiel.prototype.requestRemove = function(
    objListenpunkt      // Zu entfernenden Listenpunkt
) {
    if ((objListenpunkt.liste == this.listAnswere) && !this.needRestart) {
        if (!this.data.multiple) {
            var clone = this.listAvailable.add(objListenpunkt.index);
            clone.animate(objListenpunkt.getTop());
        }
        objListenpunkt.remove();
    }
}

/**
* Die korrekte Lösung ermitteln
* @return Array     Feld mit Indizes
*/
Reihenfolgenspiel.prototype.getSolutionIndizes = function() {
    var solution = [];
    var pos;
    for (var index = 0; index < this.data.parts.length; ++index) {
        if ((pos = this.data.parts[index].index) !== undefined) {
            if (!(pos instanceof Array))
                pos = [ pos ];
            pos.map(function(pos) {
                while (solution.length <= pos)
                    solution.push(undefined);
                solution[pos] = index;
            });
        }
    }
    return solution.filter(function(el) { return (el !== undefined); });
}

// Soll die aktive Meldung bei Veränderung der Liste gelöscht werden?
Reihenfolgenspiel.prototype._needClear = false;

/**
* Eine Meldung über das Ergebnis der Antwortüberprüfung anzeigen und ggf. die Weiterbearbeitung deaktivieren
* @return undefined
*/
Reihenfolgenspiel.prototype.reportResult = function(
    mixedMsg,   // Key für this.data, unter dem der Antworttext zu finden ist, oder Antworttext
    isKey,      // Ist mixedMsg ein Key?
    boolSuccess // Optional: War die Antwort korrekt?
) {
    if (isKey)
        mixedMsg = this.data[mixedMsg] ? this.data[mixedMsg] : (boolSuccess ? 'Richtig' : 'Falsch');
    this.showMsg(mixedMsg, boolSuccess ? ScoTools.MSGTYPE_SUCCESS : ScoTools.MSGTYPE_FAILED);
    if (this.data.forceRestart) {
        this.needRestart = true;
    } else {
        this._needClear = true;
        this._setSubmitLabel(this.data.labelReedit ? this.data.labelReedit : 'Antwort verändern');
    }
    if (boolSuccess && !this.getDone() && this.canSave())
        this.showMsg('Aufgabe wurde bereits früher erfolgreich erledigt', ScoTools.MSGTYPE_SUCCESS, true);
    if (this.canSave()) {
        var score = this.read('score');
        score = score ? score : 0;
        if (boolSuccess)
            this.saveScore(Math.max(0, 100 - score), 0, 100, 'passed');
        else
            this.saveScore(Math.min(51, score + 1), 0, 100, 'incomplete');
    }
    this.reportTime();
    if (boolSuccess) {
        this.listAnswere.keyEmpty = 'emptyRedo';
    }
    if (this.ready)
        this.playAudio(boolSuccess);
}

/**
* Den n-ten Listenpunkt in this.listAnswere ggf. als richtig (correctness = true), falsch (correctness = false) oder unbekannt (correctness = undefined) kennzeichnen
* @return undefined
*/
Reihenfolgenspiel.prototype.markCorrectness = function(
    n,              // Optional: Index der fraglichen Listenpunkt-Instanz in this.listAvailable (Standardwert: Alle Listenpunkte)
    correctness     // Ist der Listenpunkt richtig (true), falsch (false) oder noch nicht geprüft (undefined)?
) {
    if (this.data.highlightError)
        this.listAnswere.markCorrectness(n, correctness);
    else
        this.listAnswere.update(n);
}

/**
* Antwort überprüfen und Rückmeldung geben, die nicht zu viel über tatsächliche Lösung verrät
* @return Boolean       Ist die Antwort korrekt?
*/
Reihenfolgenspiel.prototype._isCorrectIntransparent = function() {
    var answere = this.listAnswere.getIndizes();
    var solution = this.getSolutionIndizes();
    if (solution.length == answere.length) {
        if (solution.reduce(function(prev, el, index) {
            return prev && (el == answere[index]);
        }, true)) {
            this.reportResult('responseCorrect', true, true);
            this.markCorrectness(undefined, true);
            return true;
        }
    }
    this.markCorrectness();
    var n;
    // Verbotener Textbaustein?
    var wrongIndex = false;
    var badness = 0;
    var thisBadness;
    for (n = 0; n < answere.length; ++n) {
        if (solution.indexOf(answere[n]) != -1)
            continue;
        thisBadness = (this.data.parts[answere[n]].badness === undefined) ? (this.data.parts[answere[n]].response ? 0.5 : 0) : this.data.parts[answere[n]].badness;
        if (thisBadness > badness) {
            wrongIndex = n;
            badness = thisBadness;
        }
    }
    if (badness) {
        if (this.data.parts[answere[wrongIndex]].response)
            this.reportResult(this.data.parts[answere[wrongIndex]].response, false);
        else
            this.reportResult('responseForbiddenParts', true);
        this.markCorrectness(wrongIndex, false);
        return false;
    }
    // Fehlt ein Textbaustein?
    for (n = 0; n < solution.length; ++n) {
        if (answere.indexOf(solution[n]) != -1)
            continue;
        wrongIndex = true;
        if (this.data.parts[solution[n]].response) {
            this.reportResult(this.data.parts[solution[n]].response, false);
            return false;
        }
    }
    if (wrongIndex) {
        this.reportResult('responseMissing', true);
        return false;
    }
    // Sonstigen Fehler finden
    for (n = 0; n < answere.length; ++n) {
        if (n >= solution.length) {
            this.reportResult('responseTooLong', true);
            this.markCorrectness(n, false);
            return false;
        }
        if (answere[n] != solution[n]) {
            this.reportResult(this.data.highlightError ? 'responseMisplacedHighlighted' : 'responseMisplaced', true);
            this.markCorrectness(n, false);
            return false;
        }
    }
    // Verwendete Textbausteine müssten erneut wiederholt werden
    this.reportResult('responseTooShort', true);
    return false;
}

/**
* Antwort überprüfen und möglichst aussagekräftige Rückmeldung geben
* @return Boolean       Ist die Antwort korrekt?
*/
Reihenfolgenspiel.prototype._isCorrectTransparent = function() {
    var answere = this.listAnswere.getIndizes();
    var solution = this.getSolutionIndizes();
    this.markCorrectness();
    var n;
    // Nach schwerem Fehler suchen: Verbotener Textbaustein?
    var wrongIndex = false;
    var badness = 0;
    var thisBadness;
    for (n = 0; n < answere.length; ++n) {
        if (solution.indexOf(answere[n]) != -1)
            continue;
        thisBadness = (this.data.parts[answere[n]].badness === undefined) ? (this.data.parts[answere[n]].response ? 0.5 : 0) : this.data.parts[answere[n]].badness;
        if (thisBadness > badness) {
            wrongIndex = n;
            badness = thisBadness;
        }
    }
    if (badness) {
        if (this.data.parts[answere[wrongIndex]].response)
            this.reportResult(this.data.parts[answere[wrongIndex]].response, false);
        else
            this.reportResult('responseForbiddenParts', true);
        this.markCorrectness(wrongIndex, false);
        return false;
    }
    // Ansonsten von oben nach unten durchkorrigieren
    for (n = 0; n < solution.length; ++n) {
        // Am Ende fehlt noch was?
        if (n >= answere.length) {
            if (this.data.parts[solution[n]].response)
                this.reportResult(this.data.parts[solution[n]].response, false);
            else if (answere.indexOf(solution[n]) == -1)
                this.reportResult('responseMissing', true);
            else
                this.reportResult('responseTooShort', true);
            return false;
        }
        // Der Eintrag ist hier falsch?
        if (solution[n] != answere[n]) {
            this.markCorrectness(n, false);
            if (solution.indexOf(answere[n]) == -1) {
                if (this.data.parts[answere[n]].response)
                    this.reportResult(this.data.parts[answere[n]].response, false);
                else
                    this.reportResult('responseForbiddenParts', true);
            } else
                this.reportResult(this.data.highlightError ? 'responseMisplacedHighlighted' : 'responseMisplaced', true);
            return false;
        }
        // Der Eintrag passt hier?
        this.markCorrectness(n, true);
    }
    // Unnötige Einträge am Ende?
    if (answere.length > solution.length) {
        this.markCorrectness(solution.length, false);
        if (this.data.parts[answere[solution.length]].response)
            this.reportResult(this.data.parts[answere[solution.length]].response, false);
        else if (solution.indexOf(answere[solution.length]) == -1)
            this.reportResult('responseWrong', true);
        else
            this.reportResult('responseTooLong', true);
        return false;
    }
    this.reportResult('responseCorrect', true, true);
    return true;
}

/**
* Antwort überprüfen und Rückmeldung geben
* @return Boolean       Ist die Antwort korrekt?
*/
Reihenfolgenspiel.prototype.isCorrect = function() {
    return (this.data.transparent ? this._isCorrectTransparent() : this._isCorrectIntransparent());
}
